'''
Generate SQL for indexia database oprerations.

'''
class Inquiry:
    '''
    Generate SQL strings from dynamic inputs. 
    
    '''
    def create(tablename, columns):
        '''
        Get a SQL CREATE TABLE statement.

        Parameters
        ----------
        tablename : str
            Name of the table to create.
        columns : dict(str)
            Dict of columns to add to table. Keys are 
            column names, values are data types.

        Returns
        -------
        create : str
            A formatted SQL CREATE TABLE statement.

        '''
        columns = [f'{col} {dtype}' for col, dtype in columns.items()]
        columns = ','.join(columns)
        create = f'CREATE TABLE IF NOT EXISTS {tablename}'
        create = f'{create} ({columns})'
        
        return create
    
    def insert(tablename, values, columns=None):
        '''
        GET a SQL INSERT statement.

        Parameters
        ----------
        tablename : str
            Name of table into which values will be inserted.
        values : list(str) or list(tuple(str))
            A list of strings or tuples containing strings. 
            Should be equal-length values representing the 
            values to insert.

        Returns
        -------
        insert : str
            A formatted SQL INSERT statement.

        '''
        values = ','.join(
            '(' + ','.join(f"'{v[i]}'" for i,_ in enumerate(v)) + ')' 
            for v in values
        )
        
        columns = f" ({','.join(columns)})" if columns else ''
        insert = f'INSERT INTO {tablename}{columns}'
        insert = f'{insert} VALUES {values}'
        
        return insert
    
    def select(tablename, columns, conditions=''):
        '''
        GET a SQL SELECT statement.

        Parameters
        ----------
        tablename : str
            Name of the table from which to select values.
        columns : list(str)
            list of column names to select.
        conditions : str, optional
            A SQL-formatted string of conditions. The default is ''.

        Returns
        -------
        select : str
            A formatted SQL SELECT statement.

        '''
        columns = ','.join(columns)
        select = f'SELECT {columns} FROM {tablename} {conditions}'
        
        return select
    
    def delete(tablename, conditions=''):
        '''
        Get a SQL DELETE FROM statement.

        Parameters
        ----------
        tablename : str
            Name of the table from which to delete.
        conditions : str, optional
            Optional WHERE conditions. The default is ''.

        Returns
        -------
        delete : str
            A formatted SQL DELETE FROM statement.

        '''
        delete = f'DELETE FROM {tablename} {conditions}'
        
        return delete
    
    def update(tablename, set_cols, set_values, conditions=''):
        '''
        Get a SQL UPDATE statement.

        Parameters
        ----------
        tablename : str
            Name of the table in which to update rows.
        set_cols : list(str)
            List of column names to update.
        set_values : list(any)
            List of values with which to update columns. Paired with 
            set_cols such that set_cols[i] = set_values[i].
        conditions : str, optional
            A SQL-formatted string of conditions. The default is ''.

        Returns
        -------
        update : TYPE
            DESCRIPTION.

        '''
        set_text = ''
        
        for i, _ in enumerate(set_cols):
            set_text += f"{set_cols[i]} = '{set_values[i]}'"
            
        update = f'UPDATE {tablename} SET {set_text} {conditions}'
        
        return update
    
    def where(cols, vals, conjunction='AND'):
        '''
        Construct WHERE condition from columns & values

        Parameters
        ----------
        cols : list(str)
            List of column names.
        vals : list(any)
            List of values.
        conjunction : str, optional
            SQL keyword to use as conjunction between 
            clauses (e.g., AND, OR).

        Returns
        -------
        conditions : str
            A SQL-formatted WHERE condition.

        '''
        where = f"WHERE {cols[0]} = '{vals[0]}' "
        
        where += ' '.join([
            f"{conjunction} {cols[i]} = '{vals[i]}'" for i in range(1, len(cols))
        ])
        
        return where


class Tabula:
    '''
    Defines columns & data types of indexia tables.
    
    '''
    def get_creator_table(genus, trait):
        '''
        Get name & columns of a creator (parent) table.

        Parameters
        ----------
        genus : str
            Name of the creator (parent) table.
        trait : str
            Name of the creator's text attribute.

        Returns
        -------
        creator_table : tuple(str, dict)
            A tuple whose first entry is the name of the creator table, 
            & whose second is a dict of table columns & data types.

        '''
        creator_table = (genus, {
            'id': 'INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL',
            trait: 'TEXT UNIQUE NOT NULL'
        })
        
        return creator_table
    
    def get_creature_table(creator, species, trait):
        '''
        Get name & columns of a creature (child) table.

        Parameters
        ----------
        creator : str
            Name of the creator (parent) table.
        name : str
            Name of the creature table.
        attribute : str
            Name of the creature's text attribute.

        Returns
        -------
        creature_table : tuple(str, dict)
            A tuple whose first entry is the name of the creature table,
            & whose second is a dict of table columns & data types.

        '''
        creature_table = (species, {
            'id': 'INTEGER PRIMARY KEY AUTOINCREMENT',
            trait: 'TEXT NOT NULL',
            f'{creator}_id': 'INTEGER NOT NULL',
            f'FOREIGN KEY ({creator}_id)': Tabula.references(creator, 'id')
        })
        
        return creature_table
    
    def references(
            tablename, on_column, on_delete='CASCADE', on_update='CASCADE'
        ):
        '''
        Generate SQL-formatted REFERENCES clause.

        Parameters
        ----------
        tablename : str
            Name of the referenced table.
        on_column : str
            Name of the referenced column.
        on_delete : str, optional
            Behavior of the child entity when the parent 
            entity is deleted. The default is 'CASCADE'.
        on_update : str, optional
            Behavior of the child entity when the parent 
            entity is updated. The default is 'CASCADE'.

        Returns
        -------
        references : str
            A SQL-formatted REFERENCES clause.

        '''
        references = f'REFERENCES {tablename}({on_column})'
        references = f'{references} ON DELETE {on_delete}'
        references = f'{references} ON UPDATE {on_update}'
        
        return references


'''
Defines tree & graph representations of indexia data.

'''
from indexia.indexia import Indexia
from pyvis.network import Network
import itertools
import networkx as nx
import os
import pandas as pd
import time
import webbrowser
import xml.etree.ElementTree as et


class ScalaNaturae:
    '''
    Ascend & descend the hierarchy of indexia data.
    
    '''
    def __init__(self, db):
        '''
        Creates a ScalaNaturae instance.

        Parameters
        ----------
        db : str
            Path to the indexia database file.

        Returns
        -------
        None.

        '''
        self.db = db
    
    def upward(self, species, creature):
        '''
        Climb up one rung.

        Parameters
        ----------
        species : str
            Name of the starting creature table.
        creature : pandas.DataFrame
            A single-row dataframe of creature entity data.

        Returns
        -------
        next_rung : list(tuple)
            List containing one tuple of the form (genus, creator),
            where genus is the name of the creator table & creator 
            is a single-row dataframe of creator entity data.

        '''
        with Indexia(self.db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            next_rung = ix.get_creator(cnxn, species, creature)
             
        return next_rung
     
    def downward(self, genus, creator):
        '''
        Climb down one rung.

        Parameters
        ----------
        genus : str
            Name of the starting creator table.
        creator : pandas.DataFrame
            A single-row dataframe of creator entity data.

        Returns
        -------
        next_rung : list(tuple)
            List of tuples of the form (species, creature), 
            where species is the name of the creature table 
            & creature is a dataframe of creature entity data.

        '''
        with Indexia(self.db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            next_rung = ix.get_creatures(cnxn, genus, creator)
            
        return next_rung
     
    def climb(self, kind, being, direction):
        '''
        Climb one rung in either direction (up or down).

        Parameters
        ----------
        kind : str
            Name of the starting table.
        being : pandas.DataFrame
            Dataframe of creator or creature entities. If the 
            dataframe contains more than one row, only results 
            for the first row will be returned.
        direction : str
            Direction to climb. Must be either 'up' or 'down'.

        Raises
        ------
        ValueError
            If direction is not either 'up' or 'down', rasise 
            a ValueError.

        Returns
        -------
        next_rung : list(tuple)
            List of tuples of the form (kind, beings), where 
            kind is the name of a creator or creature table,
            & beings is a dataframe of creator or creature 
            entity data.

        '''
        if direction == 'up':
            next_rung = self.upward(kind, being)
        elif direction == 'down':
            next_rung = self.downward(kind, being)
        else:
            raise ValueError('climb direction must be "up" or "down".')
            
        return next_rung


class Dendron:
    '''
    Represent indexia data as an XML tree.
    
    '''
    def __init__(self, db):
        '''
        Creates a Dendron instance.
        
        Sets the trunk attribute to a ScalaNaturae instance 
        for ascending & descending the hierarchy of indexia 
        data.

        Parameters
        ----------
        db : str
            Path to the indexia database.

        Returns
        -------
        None.

        '''
        self.db = db
        self.trunk = ScalaNaturae(self.db)
        
    def render_image(self, genus, creators, root=et.Element('root')):
        '''
        Render the XML tree.

        Parameters
        ----------
        genus : str
            Name of the top-level table.
        creators : pandas.DataFrame
            One or more rows of the top-level table to 
            render as XML.
        root : xml.etree.ElementTree.Element, optional
            Root element of the XML tree, used in iterative 
            calls to this method. It is not typically 
            necessary to supply this argument. The default 
            is xml.etree.ElementTree.Element('root').

        Returns
        -------
        image : xml.etree.ElementTree.ElementTree
            An XML element tree of indexia data.

        '''
        for i, creator in creators.iterrows():
            attrs = {c: creator[c] for c in creators.columns}
            
            next_rung = self.trunk.downward(
                genus, pd.DataFrame(data=attrs, index=[0])
            )
            
            branch = et.SubElement(
                root, genus, attrib={a: str(attrs[a]) for a in attrs}
            )
            
            for species, creatures in next_rung:
                self.render_image(
                    species, creatures, root=branch
                )
        
        image = et.ElementTree(root)
        
        return image
            
    def write_image(self, image, file_path=None, open_browser=True):
        '''
        Write the XML image of the Dendron instance to 
        an XML file, & optionally open in the browser.

        Parameters
        ----------
        image : xml.etree.ElementTree.ElementTree
            Image of the current Dendron instance as an 
            XML tree.
        file_path : str, optional
            Path where the XML file will be created. If 
            None, the default (dendron.xml) is used. The 
            default is None.
        open_browser : bool, optional
            If True, open the XML file in the default browser. 
            The default is True.

        Returns
        -------
        file_path : str
            Absolute path to the XML image file.

        '''
        file_path = file_path if file_path else 'dendron.xml'
        
        if os.path.isfile(file_path):
            os.remove(file_path)
        
        image.write(file_path)
        
        if open_browser:
            webbrowser.open(f'file://{os.path.abspath(file_path)}')
            time.sleep(2)
            
        return file_path
    

class Corpus:
    '''
    Represent indexia data as a dataframe.
    
    '''
    def __init__(self, db, genus, creators, max_depth=10):
        '''
        Creates a Corpus instance for the given creator data.
        
        Sets the spine attribute to a ScalaNaturae instance 
        for descending the hierarchy of creator data.

        Parameters
        ----------
        db : str
            Path to the indexia database file.
        genus : str
            Name of the creator (parent) table.
        creators : pandas.DataFrame
            Dataframe of creator entity data.
        max_depth : int, optional
            Maximum number of levels to descend when assembling 
            the corpus. The default is 10.

        Returns
        -------
        None.

        '''
        self.db = db
        self.genus = genus
        self.creators = creators
        self.max_depth = max_depth
        self.spine = ScalaNaturae(self.db)
    
    def get_trait(self, species, creatures):
        '''
        Gets the trait (attribute) column of the given 
        species.

        Parameters
        ----------
        species : str
            Name of the creature (child) table.
        creatures : pandas.DataFrame
            Dataframe of creature entity data.

        Raises
        ------
        ValueError
            If no trait column is identified, or if more 
            than one trait column is identified, raise a 
            ValueError.

        Returns
        -------
        trait : str
            Name of the trait column.

        '''
        columns = list(creatures.columns)
        traits = [c for c in columns if c != 'id' and not c.endswith('_id')]
        
        if not traits or len(traits) > 1:
            err_msg = 'Found multiple trait columns'
            err_msg = err_msg if traits else 'Found no trait column'
            err_msg = f'{err_msg} for {species}.'
            raise ValueError(err_msg)
            
        trait = traits[0]
        
        return trait
    
    def make_member(self, genus, creator, species, creatures):
        '''
        Creates a dataframe of indexia entity data.

        Parameters
        ----------
        genus : str
            Name of the creator (parent) table.
        creator : pandas.DataFrame
            Single-row dataframe of creator entity data.
        species : str
            Name of the creature (child) table.
        creatures : pandas.DataFrame
            Dataframe of creature entity data.

        Returns
        -------
        member : pandas.DataFrame
            Dataframe describing creature entities, including 
            creator information.

        '''
        creator_id = None if creator.empty else creator.id.values[0]
        trait = self.get_trait(species, creatures)
        member = pd.DataFrame()
        
        for i, creature in creatures.iterrows():
            member = pd.concat([member, pd.DataFrame(data={
                'genus': [genus],
                'creator_id': [creator_id],
                'species': [species],
                'creature_id': [creature['id']],
                'trait': [trait],
                'expression': [creature[trait]]
            })], axis=0)
            
        return member
    
    def make_limbs(self, genus, creator, depth):
        '''
        Moves down the spine to create lists of dataframes 
        representing indexia entity data.

        Parameters
        ----------
        genus : str
            Name of the creator (parent) table.
        creator : pandas.DataFrame
            Single-row dataframe of creator entity data.
        depth : int
            Current level in the corpus rendering process. 
            Compared with max_depth to determine whether 
            to proceed.

        Returns
        -------
        limbs : list(pandas.DataFrame)
            List of dataframes representing indexia entity 
            data.

        '''
        limbs = []
        
        if depth < self.max_depth:
            next_rung = self.spine.downward(genus, creator)
        else:
            next_rung = []        
        
        for species, creatures in next_rung:
            limbs += [self.make_member(
                genus, creator, species, creatures
            )]
            
            for i in range(creatures.shape[0]):
                limbs += self.make_limbs(
                    species, creatures.iloc[[i]], depth + 1
                )
        
        return limbs
    
    def assemble(self):
        '''
        Assemble the corpus of each of the creator entities.

        Returns
        -------
        corpus : pandas.DataFrame
            Dataframe representing all creatures of the 
            instance's creator entity, up to the distance 
            specified by max_depth.

        '''
        head = self.make_member(
            None, pd.DataFrame(), self.genus, self.creators
        )
        
        body = []
        
        for i in range(self.creators.shape[0]):
            creator = self.creators.iloc[[i]]
            
            body += [pd.concat(self.make_limbs(
                self.genus, creator, 0
            ), axis=0)]
            
        body = pd.concat(body, axis=0)
        corpus = pd.concat([head, body], axis=0)
        corpus.index = [i for i in range(corpus.shape[0])]
        
        return corpus
    
    def to_csv(self, corpus, file_path, **kwargs):
        '''
        Save an assembled corpus dataframe to a CSV file.

        Parameters
        ----------
        corpus : pandas.DataFrame
            Dataframe representing indexia data, created by the 
            assemble method of this class.
        file_path : str
            Path of the CSV file to be created.
        **kwargs : any
            Any keyword arguments accepted by pandas.DataFrame.to_csv.

        Returns
        -------
        file_path : str
            Path to the corpus CSV file.

        '''
        corpus.to_csv(file_path, **kwargs)
        
        return file_path

class Diktua:
    '''
    Represent indexia data as a network graph.
    
    '''
    def __init__(self, corpus, as_nodes, as_edges, self_edges=False):
        '''
        Creates an Indexinet instance.

        Parameters
        ----------
        corpus : pandas.DataFrame
            Dataframe of indexia data to represent as 
            a network graph.
        as_nodes : str
            Name of the creator or creature attribute 
            to treat as graph nodes.
        as_edges : str
            Name of the creator or creature attribute 
            to treat as graph edges.
        self_edges : bool, optional
            Whether to allow self-edges in the graph. 
            The default is False.

        Returns
        -------
        None.

        '''
        self.corpus = corpus
        self.as_nodes = as_nodes
        self.as_edges = as_edges
        self.self_edges = self_edges
        self.G = self.make_undirected_graph()
        
    def get_graph_elements(self):
        '''
        Get graph nodes & edges.

        Returns
        -------
        nodes : list
            List of graph nodes.
        edges : list
            List of tuples representing graph edges.

        '''
        sharing_nodes = list(
            self.corpus.groupby(self.as_edges).groups.values()
        )
        
        get_nodes = lambda index_list: list(
            self.corpus.loc[index_list][self.as_nodes]
        )
        
        edges = set()
        
        for indices in sharing_nodes:
            node_edges = [tuple(sorted(c)) for c in list(
                itertools.combinations(get_nodes(indices), 2)
            )]
                        
            if not self.self_edges:
                node_edges = [e for e in node_edges if e[0] != e[1]]
                
            edges = edges.union(set(node_edges))
        
        nodes = list(set(e for edge in edges for e in edge))
        edges = list(edges)
                
        return nodes, edges
    
    def make_undirected_graph(self):
        '''
        Create an undirected network graph from 
        the corpus attribute of the instance.

        Returns
        -------
        G : networkx.Graph
            And undirected network graph of 
            instance data.

        '''
        nodes, edges = self.get_graph_elements()
        G = nx.Graph()
        G.add_nodes_from(nodes)
        G.add_edges_from(edges)
        
        return G
    
    def get_node_info(self):
        '''
        Count node edges & assign titles.
        
        Edge counts are used to determine node size 
        when the graph is displayed; titles are shown when 
        hovering over nodes in the display.

        Returns
        -------
        node_edges : dict
            Keys are graph nodes; values are counts 
            of edges on each node.
        node_titles : dict
            Keys are graph nodes; values are string 
            titles assigned to nodes.

        '''
        node_edges = {}
        node_titles = {}

        for _, adjacencies in enumerate(self.G.adjacency()):
            node, adj = adjacencies
            num_edges = len(adj)
            node_edges[node] = num_edges
            node_titles[node] = f'({num_edges})'

        return node_edges, node_titles
    
    def get_node_sizes(self, node_edges, min_size, max_size):
        '''
        Calculate node size based on number of edges.
        
        Node sizes are scaled to the interval [min_size, max_size].

        Parameters
        ----------
        node_edges : dict
            Dictionary of graph nodes & edge counts.
        min_size : int
            Minimum node size.
        max_size : int
            Maximum node size.

        Returns
        -------
        node_sizes : dict
            Keys are graph nodes; values are node sizes.

        '''
        max_edges = max(node_edges.values())
        offset = max_size - min_size
        node_sizes = {}
            
        for n in node_edges:
            node_size = min_size + round(
                (offset * (node_edges[n] / max_edges))
            )
            
            node_sizes[n] = node_size
            
        return node_sizes
    
    def style_nodes(self, min_size=7, max_size=49):
        '''
        Set size & title attributes of graph nodes.

        Parameters
        ----------
        min_size : int, optional
            Minimum node size. The default is 7.
        max_size : int, optional
            Maximum node size. The default is 49.

        Returns
        -------
        networkx.Graph
            Network graph with node attributes set.

        '''
        node_edges, node_titles = self.get_node_info()
        node_sizes = self.get_node_sizes(node_edges, min_size, max_size)
        nx.set_node_attributes(self.G, node_sizes, 'size')
        nx.set_node_attributes(self.G, node_titles, 'title')
        
        return self.G

    def plot(self, plot_path=None, open_browser=False):
        '''
        Create a plot of the instance's graph.

        Parameters
        ----------
        plot_path : str or None, optional
            If supplied, plot will be written to an 
            HTML file at plot_path. The default is 
            None.
        open_browser : bool, optional
            Whether to open the plot in the browser. 
            The default is False.

        Returns
        -------
        plot : pyvis.network.Network
            A plot of the instance's network graph.
        plot_path : str or None
            If plot_path is set, returns the path of 
            the output HTML file. Otherwise None.

        '''
        plot = Network(select_menu=True, filter_menu=True)
        plot.from_nx(self.G)
        plot.show_buttons()
        
        if plot_path:
            plot.write_html(plot_path, open_browser=open_browser)
        
        return plot, plot_path
        
    def to_csv(self, file_path, **kwargs):
        '''
        Save the edges of the instance's graph to a CSV file 
        with columns 'source' & 'target'.

        Parameters
        ----------
        file_path : str
            Path of the CSV file to be created.
        **kwargs : any
            Any keyword arguments accepted by pandas.DataFrame.to_csv.

        Returns
        -------
        file_path : str
            Path to the output CSV file.

        '''
        edges = pd.DataFrame(data={
            'source': [e[0] for e in self.G.edges],
            'target': [e[1] for e in self.G.edges]
        })
        
        edges.to_csv(file_path, **kwargs)
        
        return file_path
    


'''
Make sample creators & creatures.

'''
from indexia.indexia import Indexia
from datetime import datetime as dt, timedelta as td
import random


class Maker:
    '''
    Fashion any number of creators & creatures for testing.
    
    '''
    def __init__(self, test_db, species_per_genus, num_beings, trait):
        '''
        Create a Maker instance.

        Parameters
        ----------
        test_db : str
            Path to a test database file.
        species_per_genus : int
            Number of creature tables to be added to the 
            genus.
        num_beings : int
            Number of creature records to be created.
        trait : str
            Text attribute of the creatures.

        Returns
        -------
        None.

        '''
        self.test_db = test_db
        self.species_per_genus = species_per_genus
        self.num_beings = num_beings
        self.trait = trait
        
    def make_creators(self, ix, cnxn, genus):
        '''
        Make creator beings.
        
        The number of beings specified by num_beings
        will be created in a table named genus.

        Parameters
        ----------
        ix : indexia.indexia.Indexia
            An Indexia instance.
        cnxn : sqlite3.Connection
            A database connection.
        genus : str
            Name of the creator (parent) table.

        Returns
        -------
        creators : list(pandas.DataFrame)
            Dataframe of creator data.

        '''
        for i in range(self.num_beings):
            ix.add_creator(
                cnxn, genus, 
                self.trait, f'{genus}_{i}'
            )
            
        sql = 'SELECT * FROM creators;'
        creators = ix.get_df(cnxn, sql)
        
        return creators
        
    def make_creatures(self, ix, cnxn, genus, species):
        '''
        Make sample creatures with a given genus & species.
        
        A creature table named species is created, & num_beings 
        creature records are added to the table. Each creature 
        has the attribute trait.
    
        Parameters
        ----------
        ix : indexia.indexia.Indexia
            An Indexia instance.
        cnxn : sqlite3.Connection
            A database connection.
        genus : str
            Name of the creator (parent) table.
        species : str
            Name of the creature (child) table.
    
        Returns
        -------
        creatures : pandas.DataFrame
            Dataframe of creature data.
    
        '''
        for i in range(self.num_beings):
            creator = ix.get_by_id(cnxn, genus, i + 1)
            
            ix.add_creature(
                cnxn, genus, creator, 
                species, self.trait, f'{species}_{i}'
            )
        
        sql = f'SELECT * FROM {species};'
        creatures = ix.get_df(cnxn, sql)
        
        return creatures
    
    def make_species(self, ix, cnxn, genus, species_prefix):
        '''
        Make one or more species of a given genus.
        
        The number of species created for in the genus is 
        given by species_per_genus. For each species created,
        num_beings creature records are added to the species, 
        each having the attribute trait.
    
        Parameters
        ----------
        ix : indexia.indexia.Indexia
            An Indexia instance.
        cnxn : sqlite3.Connection
            A database connection.
        genus : str
            Name of the creator (parent) table.
        species_prefix : str
            Prefix of the creature (child) table names.
        species : str
            Name of the creature (child) table.
    
        Returns
        -------
        species : list(pandas.DataFrame)
            List of dataframes containing creature data.
    
        '''
        species = []
        
        for i in range(self.species_per_genus):
            species_name = f'{species_prefix}_{i}'
            species += [self.make_creatures(ix, cnxn, genus, species_name)]
            
        return species
        
    def make(self):
        '''
        Make test data.
        
        Makes a single creator table & 3 generations of 
        creature tables. Each generation has species_per_genus 
        tables, with num_beings creatures in each table. The 
        creators & creatures all have the attribute trait.
    
        Returns
        -------
        fathers : list(pandas.DataFrame)
            List containing a single dataframe of creator 
            data.
        sons : list(pandas.DataFrame)
            List containing species_per_genus dataframes 
            of creature data.
        grandsons : list(pandas.DataFrame)
            List containing (species_per_genus)^2 dataframes 
            of creature data.
        great_grandsons : list(pandas.DataFrame)
            List containing (species_per_genus)^3 dataframes 
            of creature data.

        '''
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            genus = 'creators'
            fathers = [self.make_creators(ix, cnxn, genus)]
            species_prefix = 'creatures'
            sons = self.make_species(ix, cnxn, genus, species_prefix)
            grandsons = []
            
            for i in range(self.species_per_genus):
                genus = f'creatures_{i}'
                species_prefix = f'creatures_{i}'
                
                grandsons += self. make_species(
                    ix, cnxn, genus, species_prefix
                )
                
            great_grandsons = []
            
            for i in range(self.species_per_genus):
                for j in range(self.species_per_genus):
                    genus = f'creatures_{i}_{j}'
                    species_prefix = f'creatures_{i}_{j}'
                    
                    great_grandsons += self.make_species(
                        ix, cnxn, genus, species_prefix
                    )
                                
        return fathers, sons, grandsons, great_grandsons
    
    def get(self):
        '''
        Get test data.

        Returns
        -------
        fathers : list(pandas.DataFrame)
            List containing a single dataframe of creator 
            data.
        sons : list(pandas.DataFrame)
            List containing species_per_genus dataframes 
            of creature data.
        grandsons : list(pandas.DataFrame)
            List containing (species_per_genus)^2 dataframes 
            of creature data.
        great_grandsons : list(pandas.DataFrame)
            List containing (species_per_genus)^3 dataframes 
            of creature data.

        '''
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            sql = 'SELECT * FROM creators;'
            fathers = [ix.get_df(cnxn, sql)]
            
            sons = []
            grandsons = []
            great_grandsons = []
            
            for i in range(self.species_per_genus):
                sql = f'SELECT * FROM creatures_{i};'
                sons += [ix.get_df(cnxn, sql)]
                
                for j in range(self.species_per_genus):
                    sql = f'SELECT * FROM creatures_{i}_{j};'
                    grandsons += [ix.get_df(cnxn, sql)]
                    
                    for k in range(self.species_per_genus):
                        sql = f'SELECT * FROM creatures_{i}_{j}_{k};'
                        great_grandsons += [ix.get_df(cnxn, sql)]
                        
            return fathers, sons, grandsons, great_grandsons


class Templates:
    '''
    Create template indexia objects.
    
    '''
    def __init__(self, db):
        '''
        Creates a Templates instance.

        Parameters
        ----------
        db : str
            Path to the indexia database file.

        Returns
        -------
        None.

        '''
        self.db = db
        
    def show_templates(self):
        '''
        Show available templates

        Returns
        -------
        templates : dict
            Dictionary of available template names & table structures.

        '''
        templates = {
            'philosophy': {'philosophers': {'works': {'topics'}}},
            'zettelkasten': {'scribes': {'libraries': {'cards': {'keywords'}}}}
        }
        
        print('Available templates:')
        [print(f'\n{t}:', templates[t], sep='\n\t') for t in templates]
        
        return templates
        
    def build_template(self, template_name):
        '''
        Create objects for the given template.

        Parameters
        ----------
        template_name : str
            Name of the template to build.

        Raises
        ------
        ValueError
            If template_name is not a valid template, raise 
            a ValueError.

        Returns
        -------
        objects : list(tuple(string, pandas.DataFrame))
            List of tuples containing table names & object 
            data.

        '''
        objects = {}
        
        if template_name == 'philosophy':
            with Indexia(self.db) as ix:
                cnxn = ix.open_cnxn(ix.db)
                
                plato = ix.add_creator(
                    cnxn, 'philosophers', 'name', 'Plato'
                )
                
                aristotle = ix.add_creator(
                    cnxn, 'philosophers', 'name', 'Aristotle'
                )
                
                apology = ix.add_creature(
                    cnxn, 'philosophers', plato, 'works', 
                    'title', 'Apology of Socrates'
                )
                
                symposium = ix.add_creature(
                    cnxn, 'philosophers', plato, 
                    'works', 'title', 'Symposium'
                )
                
                republic = ix.add_creature(
                    cnxn, 'philosophers', plato, 
                    'works', 'title', 'Republic'
                )
                
                on_the_heavens = ix.add_creature(
                    cnxn, 'philosophers', aristotle, 
                    'works', 'title', 'On the Heavens'
                )
                
                topics = ix.add_creature(
                    cnxn, 'philosophers', aristotle, 
                    'works', 'title', 'Topics'
                )
                
                on_the_soul = ix.add_creature(
                    cnxn, 'philosophers', aristotle, 
                    'works', 'title', 'On the Soul'
                )
                
                ix.add_creature(
                    cnxn, 'works', apology, 
                    'topics', 'name', 'civics'
                )
                
                ix.add_creature(
                    cnxn, 'works', symposium, 
                    'topics', 'name', 'love'
                )
                
                ix.add_creature(
                    cnxn, 'works', republic, 
                    'topics', 'name', 'civics'
                )
                
                ix.add_creature(
                    cnxn, 'works', on_the_heavens, 
                    'topics', 'name', 'cosmology'
                )
                
                ix.add_creature(
                    cnxn, 'works', topics, 
                    'topics', 'name', 'logic'
                )
                
                ix.add_creature(
                    cnxn, 'works', on_the_soul, 
                    'topics', 'name', 'psychology'
                )
                
                ix.add_creature(
                    cnxn, 'works', apology, 
                    'topics', 'name', 'civics'
                )
                
                philosophers = ix.get_df(cnxn, 'SELECT * FROM philosophers;')
                objects['philosophers'] = philosophers
                
                works = ix.get_df(cnxn, 'SELECT * FROM works;')
                objects['works'] = works
                
                topics = ix.get_df(cnxn, 'SELECT * FROM topics;')
                objects['topics'] = topics
                
        elif template_name == 'zettelkasten':
            with Indexia(self.db) as ix:
                cnxn = ix.open_cnxn(ix.db)
                
                scribe = ix.add_creator(cnxn, 'scribes', 'name', 'Grammateus')
                
                first = ix.add_creature(
                    cnxn, 'scribes', scribe, 'libraries', 'name', 'First'
                )
                
                second = ix.add_creature(
                    cnxn, 'scribes', scribe, 'libraries', 'name', 'Second'
                )
                
                now = dt.now()
                
                keywords = [
                    'writing', 'inscription', 'zettelkasten',
                    'mnemotechnic', 'topic', 'category',
                    'antinomy', 'reason', 'anatomy',
                    'science', 'logic', 'apodeictic'
                ]
                
                for library in [first, second]:
                    for i in range(1, 4):
                        created = now + td(minutes=1)
                        created = created.strftime('%Y-%m-%d-%H-%M')
                        
                        card = ix.add_creature(
                            cnxn, 'libraries', library, 
                            'cards', 'created', created
                        )
                        
                        for keyword in random.sample(keywords, 3):
                            ix.add_creature(
                                cnxn, 'cards', card,
                                'keywords', 'keyword', keyword
                            )
                
                scribes = ix.get_df(cnxn, 'SELECT * FROM scribes;')
                objects['scribes'] = scribes
                
                libraries = ix.get_df(cnxn, 'SELECT * FROM libraries;')
                objects['libraries'] = libraries
                
                cards = ix.get_df(cnxn, 'SELECT * FROM cards;')
                objects['cards'] = cards
                
                keywords = ix.get_df(cnxn, 'SELECT * FROM keywords;')
                objects['keywords'] = keywords
                
        else:
            self.show_templates()
            raise ValueError(f'Found no template {template_name}.')
            
        return objects


from indexia.eidola import Maker
from indexia.indexia import Indexia
from indexia.schemata import Corpus, Dendron, Diktua, ScalaNaturae
from pyvis.network import Network
import itertools
import os
import pandas as pd
import unittest as ut
import xml.etree.ElementTree as et


class TestScalaNaturae(ut.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db = 'tests/data/test_schemata.db'
        cls.ladder = ScalaNaturae(cls.test_db)
        cls.species_per_genus = 3
        cls.num_beings = 5
        cls.trait = 'name'
        
        cls.maker = Maker(
            cls.test_db, cls.species_per_genus, 
            cls.num_beings, cls.trait
        )
        
        (
            cls.fathers, cls.sons, 
            cls.grandsons, cls.great_grandsons
        ) = cls.maker.get() if cls.checkEidolaExist() else cls.maker.make()
            
        
    @classmethod
    def checkEidolaExist(cls):
        exp_creator = 'creators'
        right = cls.species_per_genus - 1
        exp_creature = f'creatures_{right}_{right}_{right}'
        
        with Indexia(cls.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            tables = ix.get_all_tables(cnxn)
            eidola_exist = exp_creator in tables and exp_creature in tables
            
        return eidola_exist
        
    def testUpward(self):
        species = 'creatures_0_0_0'
        creature = self.great_grandsons[0].sample(1)
        next_rung = self.ladder.upward(species, creature)
        
        genus, creator = next_rung[0]
        exp_genus = 'creatures_0_0'
        exp_id = creature[f'{exp_genus}_id'].max()
        self.assertEqual(genus, exp_genus)
        self.assertEqual(creator.id.max(), exp_id)
        
        creator = self.fathers[0].sample(1)
        genus = 'creators'
        exp_empty = self.ladder.upward(genus, creator)
        self.assertEqual(len(exp_empty), 0)
    
    def testDownward(self):
        species = 'creatures_0_0_0'
        creature = self.great_grandsons[0].sample(1)
        next_rung_up = self.ladder.upward(species, creature)
        
        genus, creator_data = next_rung_up[0]
        next_rung_down = self.ladder.downward(genus, creator_data)
        
        species_list = [n[0] for n in next_rung_down]
        self.assertIn(species, species_list)
        creature_data = next_rung_down[species_list.index(species)][1]
        self.assertIn(creature.id.values[0], list(creature_data.id))
        
        exp_empty = self.ladder.downward(species, creature)
        self.assertEqual(len(exp_empty), 0)
    
    def testClimb(self):
        species = 'creatures_0_0_0'
        creature = self.great_grandsons[0].sample(1)
        up = self.ladder.climb(species, creature, 'up')
        genus, creator = up[0]
        self.assertEqual(len(up), 1)
        self.assertEqual('creatures_0_0', genus)
        self.assertIn('id', list(creator.columns))
        self.assertIn('name', list(creator.columns))
        
        down = self.ladder.climb(genus, creator, 'down')
        species, creature = down[0]
        self.assertIn(genus, species)
        
        self.assertEqual(
            ['id', 'name', f'{genus}_id'], list(creature.columns)
        )
        
        self.assertRaises(
            ValueError, self.ladder.climb, 
            genus, creator, 'sideways'
        )
        
        
class TestDendron(ut.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db = 'tests/data/test_schemata.db'
        cls.xml_file = 'tests/data/dendron.xml'
        cls.species_per_genus = 3
        cls.num_beings = 5
        cls.trait = 'name'
        
        cls.maker = Maker(
            cls.test_db, cls.species_per_genus, 
            cls.num_beings, cls.trait
        )
        
        (
            cls.fathers, cls.sons, 
            cls.grandsons, cls.great_grandsons
        ) = cls.maker.get() if cls.checkEidolaExist() else cls.maker.make()
        
    @classmethod
    def checkEidolaExist(cls):
        exp_creator = 'creators'
        right = cls.species_per_genus - 1
        exp_creature = f'creatures_{right}_{right}_{right}'
        
        with Indexia(cls.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            tables = ix.get_all_tables(cnxn)
            eidola_exist = exp_creator in tables and exp_creature in tables
            
        return eidola_exist
    
    def testRenderImage(self):
        genus = 'creators'
        creator = self.fathers[0].query('id == 1')
        dendron = Dendron(self.test_db)
        
        rendered = dendron.render_image(genus, creator)
        self.assertIsInstance(rendered, et.ElementTree)
        exp_son = ('creatures_0', self.sons[0].query('id == 1'))
        exp_grandson = ('creatures_0_0', self.grandsons[0].query('id == 1'))
        
        exp_great_grandson = (
            'creatures_0_0_0', self.great_grandsons[0].query('id == 1')
        )
        
        exp_creatures = [exp_son, exp_grandson, exp_great_grandson]
        
        for exp_species, exp_creature in exp_creatures:
            exp_path = f".//{exp_species}[@id='{exp_creature.id.max()}']"
            creator_element = rendered.find(exp_path)
            self.assertIsInstance(creator_element, et.Element)
    
    def testWriteImage(self):
        genus = 'creators'
        creator = self.fathers[0].query('id == 1')
        dendron = Dendron(self.test_db)
        
        image = dendron.render_image(genus, creator)
        outfile = dendron.write_image(image, self.xml_file, open_browser=False)
        self.assertEqual(self.xml_file, outfile)
    
    def tearDown(self):
        try:
            os.remove(self.xml_file)
        except:
            pass
        
class TestCorpus(ut.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db = 'tests/data/test_schemata.db'
        cls.csv_path = 'tests/data/test_corpus.csv'
        cls.species_per_genus = 3
        cls.num_beings = 5
        cls.trait = 'name'
        
        cls.maker = Maker(
            cls.test_db, cls.species_per_genus, 
            cls.num_beings, cls.trait
        )
        
        (
            cls.fathers, cls.sons, 
            cls.grandsons, cls.great_grandsons
        ) = cls.maker.get() if cls.checkEidolaExist() else cls.maker.make()
        
        cls.genus = 'creators'
        cls.creators = cls.fathers[0]
        cls.corpus = Corpus(cls.test_db, cls.genus, cls.creators, max_depth=4)
        
    @classmethod
    def checkEidolaExist(cls):
        exp_creator = 'creators'
        right = cls.species_per_genus - 1
        exp_creature = f'creatures_{right}_{right}_{right}'
        
        with Indexia(cls.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            tables = ix.get_all_tables(cnxn)
            eidola_exist = exp_creator in tables and exp_creature in tables
            
        return eidola_exist
    
    @classmethod
    def make_frame(cls, columns):
        return pd.DataFrame(columns=columns)
        
    def testGetTrait(self):
        exp_pass = self.make_frame(['id', self.trait])
        trait = self.corpus.get_trait('test_species', exp_pass)
        self.assertEqual(trait, self.trait)
        
        exp_pass = self.make_frame(['id', self.trait, 'creators_id'])
        trait = self.corpus.get_trait('test_species', exp_pass)
        self.assertEqual(trait, self.trait)
        
        exp_fail = self.make_frame(['id', 'creators_id'])
        
        self.assertRaises(
            ValueError, self.corpus.get_trait, 
            'test_species', exp_fail
        )
        
        exp_fail = self.make_frame([
            'id', self.trait, 'extra_trait', 'creators_id'
        ])
        
        self.assertRaises(
            ValueError, self.corpus.get_trait, 
            'test_species', exp_fail
        )
     
    def testMakeMember(self):
        creator = self.creators.query('id == 1')
        
        members = self.corpus.make_member(
            None, pd.DataFrame(), self.genus, creator
        )
        
        exp_columns = [
            'genus', 'creator_id', 
            'species', 'creature_id', 
            'trait', 'expression'
        ]
        
        self.assertEqual(list(members.columns), exp_columns)
        self.assertEqual(members.shape[0], creator.shape[0])
        self.assertIsNone(members.genus.values[0])
        self.assertIsNone(members.creator_id.values[0])
        self.assertEqual(members.species.values[0], self.genus)
        
        self.assertEqual(
            members.creature_id.values[0], creator.id.values[0]
        )
        
        self.assertEqual(members.trait.values[0], self.trait)
        
        self.assertEqual(
            members.expression.values[0], creator[self.trait].values[0]
        )
        
    def testMakeLimb(self):
        self.corpus.max_depth = 1
        creator = self.creators.query('id == 1')
        limb = self.corpus.make_limbs(self.genus, creator, 0)
        limb = pd.concat(limb, axis=0)
        exp_genus = {self.genus}
        exp_creator_id = {1}
        exp_species = {f'creatures_{i}' for i in range(self.species_per_genus)}
        exp_trait = {self.trait}
        self.assertEqual(set(limb.genus), exp_genus)
        self.assertEqual(set(limb.creator_id), exp_creator_id)
        self.assertEqual(set(limb.species), exp_species)
        self.assertEqual(set(limb.trait), exp_trait)
        
        self.corpus.max_depth = 0
        limb = self.corpus.make_limbs(self.genus, self.creators, 0)
        self.assertFalse(limb)
        
        genus = 'creatures_0_0_0'
        creator = self.great_grandsons[0].iloc[[0]]
        limb = self.corpus.make_limbs(genus, creator, 0)
        self.assertFalse(limb)
        
        self.corpus.max_depth = 2
        limb = self.corpus.make_limbs(self.genus, self.creators, 0)
        limb = pd.concat(limb, axis=0)
        exp_species = set()
        
        for i in range(self.species_per_genus):
            exp_species = exp_species.union({f'creatures_{i}'})
            
            for j in range(self.species_per_genus):
                exp_species = exp_species.union({f'creatures_{i}_{j}'})
        
        self.assertEqual(set(limb.species), exp_species)
        
    def testAssemble(self):
        self.corpus.max_depth = 5
        corpus = self.corpus.assemble()
        exp_index = [i for i in range(corpus.shape[0])]
        self.assertEqual(list(corpus.index), exp_index)
        exp_species = {'creators'}
        
        for i in range(self.species_per_genus):
            exp_species = exp_species.union({f'creatures_{i}'})
            
            for j in range(self.species_per_genus):
                exp_species = exp_species.union({f'creatures_{i}_{j}'})
                
                for k in range(self.species_per_genus):
                    exp_species = exp_species.union({f'creatures_{i}_{j}_{k}'})
                    
        self.assertEqual(set(corpus.species), exp_species)
        
    def testToCSV(self):
        self.corpus.max_depth = 5
        corpus = self.corpus.assemble()
        file_path = self.corpus.to_csv(corpus, self.csv_path, index=False)
        self.assertEqual(file_path, self.csv_path)
    
    @classmethod
    def tearDownClass(cls):
        try:
            os.remove(cls.csv_path)
        except:
            pass
        
class TestDiktua(ut.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db = 'tests/data/test_schemata.db'
        cls.csv_path = 'tests/data/test_diktua.csv'
        cls.html_path = 'tests/data/test_diktua.html'
        cls.species_per_genus = 3
        cls.num_beings = 5
        cls.trait = 'name'
        
        cls.maker = Maker(
            cls.test_db, cls.species_per_genus, 
            cls.num_beings, cls.trait
        )
        
        (
            cls.fathers, cls.sons, 
            cls.grandsons, cls.great_grandsons
        ) = cls.maker.get() if cls.checkEidolaExist() else cls.maker.make()
        
        cls.genus = 'creators'
        cls.creators = cls.fathers[0]
        cls.corpus = Corpus(cls.test_db, cls.genus, cls.creators, 1).assemble()
        cls.self_edges = False
        
        cls.diktua = Diktua(
            cls.corpus,
            as_nodes='species', 
            as_edges='genus',
            self_edges=cls.self_edges
        )
    
    @classmethod
    def checkEidolaExist(cls):
        exp_creator = 'creators'
        right = cls.species_per_genus - 1
        exp_creature = f'creatures_{right}_{right}_{right}'
        
        with Indexia(cls.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            tables = ix.get_all_tables(cnxn)
            eidola_exist = exp_creator in tables and exp_creature in tables
            
        return eidola_exist
    
    @classmethod
    def get_expected_edges(cls, self_edges):
        sons = [f'creatures_{i}' for i in range(cls.species_per_genus)]
        
        exp_edges = {
            tuple(sorted(i)) for i in itertools.combinations(sons, 2)
        }
        
        if self_edges:
            exp_edges = exp_edges.union({(s, s) for s in sons})
            
        return exp_edges
        
    def testGetGraphElements(self):
        exp_nodes = set(self.corpus.species.unique()) - {self.genus}
        exp_edges = self.get_expected_edges(self.self_edges)
        self.diktua.self_edges = self.self_edges
        nodes, edges = self.diktua.get_graph_elements()
        edges = [tuple(sorted(e)) for e in edges]
        self.assertEqual(set(nodes), exp_nodes)
        self.assertEqual(set(edges), exp_edges)
        
        exp_edges = self.get_expected_edges(not self.self_edges)
        self.diktua.self_edges = not self.self_edges
        nodes, edges = self.diktua.get_graph_elements()
        edges = [tuple(sorted(e)) for e in edges]
        self.assertEqual(set(edges), exp_edges)
    
    def testMakeUndirectedGraph(self):
        exp_nodes = set(self.corpus.species.unique()) - {self.genus}
        exp_edges = self.get_expected_edges(self.self_edges)
        self.diktua.self_edges = self.self_edges
        G = self.diktua.make_undirected_graph()
        self.assertEqual(set(G.nodes), exp_nodes)        
        self.assertEqual(set([tuple(sorted(e)) for e in G.edges]), exp_edges)
    
    def testGetNodeInfo(self):
        node_connections, node_titles = self.diktua.get_node_info()
        exp_connections = {2}
        exp_titles = {'(2)'}
        self.assertEqual(set(node_connections.values()), exp_connections)
        self.assertEqual(set(node_titles.values()), exp_titles)
    
    def testGetNodeSizes(self):
        min_size = 7
        max_size = 49
        max_connections = 2
        
        connections, _ = self.diktua.get_node_info()
        
        node_sizes = self.diktua.get_node_sizes(
            connections, min_size, max_size
        )
        
        exp_sizes = {max_size}
        self.assertEqual(set(node_sizes.values()), exp_sizes)
        
        connections['creatures_0'] = 0
        
        node_sizes = self.diktua.get_node_sizes(
            connections, min_size, max_size
        )
        
        exp_sizes = {min_size, max_size}
        self.assertEqual(set(node_sizes.values()), exp_sizes)
        
        connections['creatures_1'] = 1
        
        node_sizes = self.diktua.get_node_sizes(
            connections, min_size, max_size
        )
        
        mid_size = min_size + round(
            (max_size - min_size) * (1 / max_connections)
        )
        
        exp_sizes = {min_size, mid_size, max_size}
        self.assertEqual(set(node_sizes.values()), exp_sizes)
    
    def testStyleNodes(self):
        min_size = 7
        max_size = 49
        max_connections = 2
        exp_title = f'({max_connections})'
        
        exp_result = {s: {
            'size': max_size, 
            'title': exp_title
        } for s in ['creatures_0', 'creatures_1', 'creatures_2']}
        
        self.diktua.style_nodes(min_size=min_size, max_size=max_size)
        result = self.diktua.G.nodes.data()
        self.assertDictEqual(dict(result), exp_result)
    
    def testPlot(self):
        plot, path = self.diktua.plot()
        self.assertIsInstance(plot, Network)
        self.assertIsNone(path)
        
        plot, path = self.diktua.plot(self.html_path)
        self.assertTrue(os.path.isfile(path))
    
    def testToCSV(self):
        csv_path = self.diktua.to_csv(self.csv_path)
        self.assertTrue(os.path.isfile(csv_path))
    
    @classmethod
    def tearDownClass(cls):
        try:
            os.remove(cls.csv_path)
            os.remove(cls.html_path)
        except:
            pass


if __name__ == '__main__':
    ut.main()


from indexia.inquiry import Inquiry, Tabula
import unittest as ut


class TestInquiry(ut.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tablename = 'users'
        cls.values = [('user1'), ('user2'), ('user3')]
        
        cls.columns = {
            'uid': 'INT PRIMARY KEY',
            'username': 'VARCHAR(28)'
        }
        
    def testCreate(self):
        statement = Inquiry.create(self.tablename, self.columns)
        
        expected = ' '.join([
            'CREATE TABLE IF NOT EXISTS users',
            '(uid INT PRIMARY KEY,username VARCHAR(28))'
        ])
        
        self.assertEqual(statement, expected)
        
    def testInsert(self):        
        statement = Inquiry.insert(
            self.tablename, 
            [(i, f'user{i}') for i in range(1, 4)]
        )
        
        expected = ' '.join([
            'INSERT INTO users VALUES',
            "('1','user1'),('2','user2'),('3','user3')"
        ])
        
        self.assertEqual(statement, expected)
        
    def testSelect(self):
        statement = Inquiry.select(
            self.tablename, 
            ['uid'], 
            'WHERE uid > 1'
        )
        
        expected = 'SELECT uid FROM users WHERE uid > 1'
        self.assertEqual(statement, expected)
    
    def testDelete(self):
        statement = Inquiry.delete(self.tablename)
        expected = 'DELETE FROM users '
        self.assertEqual(statement, expected)
        
        statement = Inquiry.delete(
            self.tablename, 
            conditions="WHERE username = 'user1'"
        )
        
        expected = "DELETE FROM users WHERE username = 'user1'"
        self.assertEqual(statement, expected)
        
    def testUpdate(self):
        statement = Inquiry.update(
            self.tablename, 
            ['username'], 
            ['user4'],
            conditions="WHERE username = 'user1'"
        )
        
        expected = ' '.join([
            "UPDATE users SET username = 'user4'",
            "WHERE username = 'user1'"
        ])
        
        self.assertEqual(statement, expected)
    
    def testWhere(self):
        statement = Inquiry.where(
            ['username', 'username'], 
            ['user1', 'user2'],
            conjunction='OR'
        )
        
        expected = "WHERE username = 'user1' OR username = 'user2'"
        self.assertEqual(statement, expected)


class TestTabula(ut.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.genus = 'scribes'
        cls.species = 'libraries'
        cls.genus_trait = 'pseudonym'
        cls.species_trait = 'libronym'
        
    def testGetCreatorTable(self):
        genus, cols = Tabula.get_creator_table(
            self.genus, self.genus_trait
        )
        
        self.assertEqual(self.genus, genus)
        self.assertEqual({'id', self.genus_trait}, set(cols.keys()))
        
    def testGetCreatureTable(self):
        species, cols = Tabula.get_creature_table(
            self.genus, self.species, self.species_trait
        )
        
        self.assertEqual(self.species, species)
        
        self.assertEqual({
            'id', 
            self.species_trait, 
            f'{self.genus}_id', 
            f'FOREIGN KEY ({self.genus}_id)'
        }, set(cols.keys()))
    
    def testReferences(self):
        references = Tabula.references(
            self.genus, 
            self.genus_trait
        )
        
        expected = ' '.join([
            f'REFERENCES {self.genus}({self.genus_trait})',
            'ON DELETE CASCADE ON UPDATE CASCADE'
        ])
        
        self.assertEqual(references, expected)


if __name__ == '__main__':
    ut.main()


from indexia.eidola import Maker, Templates
from indexia.indexia import Indexia
import os
import pandas as pd
import unittest as ut


class TestEidola(ut.TestCase):
    def setUp(self):
        self.test_db = 'tests/data/test_eidola.db'
        self.species_per_genus = 3
        self.num_beings = 10
        self.trait = 'name'
        self.genus = 'creators'
        
        self.maker = Maker(
            self.test_db, 
            self.species_per_genus, 
            self.num_beings, 
            self.trait
        )
        
    def testMakeCreators(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            creators = self.maker.make_creators(ix, cnxn, self.genus)
            exp_columns = ['id', self.trait]
            exp_expr = f'{self.genus}_0'
            self.assertEqual(list(creators.columns), exp_columns)
            self.assertEqual(creators.shape[0], self.num_beings)
            self.assertIn(exp_expr, list(creators[self.trait]))
        
    def testMakeCreatures(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            self.maker.make_creators(ix, cnxn, self.genus)
            species = 'creatures'
            
            creatures = self.maker.make_creatures(
                ix, cnxn, self.genus, species
            )
            
            exp_columns = ['id', self.trait, f'{self.genus}_id']
            exp_expr = f'{species}_0'
            exp_fk = 1
            self.assertEqual(list(creatures.columns), exp_columns)
            self.assertEqual(creatures.shape[0], self.num_beings)
            self.assertIn(exp_expr, list(creatures[self.trait]))
            self.assertIn(exp_fk, list(creatures[f'{self.genus}_id']))
            
    
    def testMakeSpecies(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            self.maker.make_creators(ix, cnxn, self.genus)
            species_prefix = 'creatures'
            
            species = self.maker.make_species(
                ix, cnxn, self.genus, species_prefix
            )
            
            self.assertEqual(len(species), self.species_per_genus)
            self.assertIsInstance(species[0], pd.DataFrame)
    
    def testMake(self):
        fathers, sons, grandsons, great_grandsons = self.maker.make()
        self.assertEqual(len(fathers), 1)
        self.assertEqual(len(sons), self.species_per_genus)
        self.assertEqual(len(grandsons), self.species_per_genus**2)
        self.assertEqual(len(great_grandsons), self.species_per_genus**3)
        
    def testGet(self):
        fathers, sons, grandsons, great_grandsons = self.maker.get()
        self.assertEqual(len(fathers), 1)
        self.assertEqual(len(sons), self.species_per_genus)
        self.assertEqual(len(grandsons), self.species_per_genus**2)
        self.assertEqual(len(great_grandsons), self.species_per_genus**3)
        
    def tearDown(self):
        try:
            os.remove(self.test_db)
        except:
            pass
        
        
class TestTemplates(ut.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_db = 'tests/data/test_eidola.db'
        cls.generator = Templates(cls.test_db)
    
    def testShowTemplates(self):
        templates = self.generator.show_templates()
        exp_templates = ['philosophy', 'zettelkasten']
        self.assertEqual(list(templates.keys()), exp_templates)
    
    def testBuildTemplate(self):
        self.assertRaises(
            ValueError, self.generator.build_template, 
            'fake_template'
        )
        
        objects = self.generator.build_template('philosophy')
        self.assertEqual(len(objects), 3)
        species = list(objects.keys())
        exp_species = ['philosophers', 'works', 'topics']
        self.assertEqual(species, exp_species)
        
        objects = self.generator.build_template('zettelkasten')
        self.assertEqual(len(objects), 4)
        species = list(objects.keys())
        exp_species = ['scribes', 'libraries', 'cards', 'keywords']
        self.assertEqual(species, exp_species)
        
    @classmethod
    def tearDownClass(cls):
        try:
            os.remove(cls.test_db)
        except:
            pass


if __name__ == '__main__':
    ut.main()



'''
Defines core operations on indexia objects.

'''
from indexia.inquiry import Inquiry, Tabula
import os
import sqlite3
import pandas as pd


class Indexia:
    '''
    Core class for creating, modifying, & retrieving 
    indexia objects.
    
    '''
    def __init__(self, db=None):
        '''
        Create an indexia instance & build a path to 
        a default database file if one is not supplied.

        Parameters
        ----------
        db : str, optional
            Path to a database file. The default is None.

        Returns
        -------
        None.

        '''
        self.cnxns = {}
        
        self.db = db if db else os.path.join(
            os.path.abspath(__file__),
            '..', 'data', 'indexia.db'
        )
    
    def __enter__(self):
        '''
        Enable with _ as _ syntax.

        '''
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        '''
        Close all database connections on exit.

        '''
        self.close_all_cnxns()
            
    def open_cnxn(self, db):
        '''
        Open a connection to a database.

        Parameters
        ----------
        db : str
            The name of the database.

        Returns
        -------
        cnxn : sqlite3.Connection
            Connection to the database.

        '''
        cnxn = sqlite3.connect(db)
        cnxn.execute('PRAGMA foreign_keys = 1')
        
        if db in self.cnxns.keys():
            self.cnxns[db] += [cnxn]
        else:
            self.cnxns[db] = [cnxn]
        
        return cnxn
    
    def close_cnxn(self, db):
        '''
        Close connections to a database.

        Parameters
        ----------
        db : str
            Path to the database file.

        Returns
        -------
        None.

        '''
        for cnxn in self.cnxns[db]:
            cnxn.close()
        
        self.cnxns[db] = []
    
    def close_all_cnxns(self):
        '''
        Close all database connections.

        Returns
        -------
        None.

        '''
        for db in self.cnxns:
            self.close_cnxn(db)
            
    def get_or_create(self, cnxn, tablename, dtype, cols, vals, retry=True):
        '''
        Get entities from an existing table, or create 
        the table & (optionally) insert them.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        tablename : str
            Name of the database table. If the table does 
            not exist, it will be created.
        dtype : dict(str, str)
            Dict of table columns & column data types.
        cols : list(str)
            Columns to be used in SELECT statement.
        vals : list(str)
            Values to be used in SELECT statement.
        retry : bool, optional
            If true & SELECT returns an empty result, 
            INSERT the specifies values & try again.
            The default is True.

        Raises
        ------
        ValueError
            Raised when no matching rows are found 
            & retry is False.

        Returns
        -------
        result : pandas.DataFrame
            A dataframe of rows matching column & 
            value criteria.

        '''
        create = Inquiry.create(tablename, dtype)
        cnxn.execute(create)
        cnxn.commit()
        
        where = Inquiry.where(cols, vals)
        select = Inquiry.select(tablename, ['*'], where)
        result = pd.read_sql(select, cnxn)
        
        if result.empty and retry:
            insert = Inquiry.insert(tablename, [tuple(vals)], columns=cols)
            cnxn.execute(insert)
            cnxn.commit()
            
            return self.get_or_create(
                cnxn, tablename, dtype, cols, vals, retry=False
            )
        
        elif result.empty:
            raise ValueError(f'No rows in {tablename} where {where}.')
        
        return result
                
    def delete(self, cnxn, species, entity_id):
        '''
        Delete an entity from a table by ID.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        species : str
            Name of the table from which to delete.
        entity_id : int
            ID of the entity to delete.

        Returns
        -------
        rows_deleted : int
            Count of rows affected by DELETE statement.

        '''
        where = Inquiry.where(['id'], [entity_id])
        delete = Inquiry.delete(species, where)
        cursor = cnxn.cursor()
        cursor.execute(delete)
        cnxn.commit()
        rows_deleted = cursor.rowcount
        
        return rows_deleted
    
    def update(
            self, cnxn, tablename, 
            set_cols, set_vals, 
            where_cols, where_vals
        ):
        '''
        Update values in a database table. Executes a SQL statement 
        of the form
        
        UPDATE 
            {tablename}
        SET 
            {set_cols[0]} = {set_vals[0]},
            {set_cols[1]} = {set_vals[1]},
            ...
        WHERE 
            {where_cols[0]} = {where_vals[0]} AND
            {where_cols[1]} = {where_vals[1]} AND
            ...

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        tablename : str
            Name of the table to update.
        set_cols : list(str)
            List of columns to update.
        set_vals : list(any)
            Updated values for columns.
        where_cols : list(str)
            List of columns for WHERE condition.
        where_vals : list(any)
            List of values for WHERE condition.

        Returns
        -------
        rows_updated : int
            Number of rows affected by update statement.

        '''
        where = Inquiry.where(where_cols, where_vals)
        update = Inquiry.update(tablename, set_cols, set_vals, where)
        cursor = cnxn.cursor()
        cursor.execute(update)
        cnxn.commit()
        rows_updated = cursor.rowcount
        
        return rows_updated
    
    
    ##########
    # adders #
    ##########
    
    def add_creator(self, cnxn, genus, trait, expr):
        '''
        Get or create a creator entity.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        genus : str
            Name of the creator (parent) table to be retrieved 
            or created.
        trait : str
            Name of the creator's text attribute.
        expr : str
            Value of the creator's text attribute.

        Returns
        -------
        creator : pandas.DataFrame
            A single-row dataframe of creator entity data.

        '''
        _, dtype = Tabula.get_creator_table(genus, trait)
        creator = self.get_or_create(cnxn, genus, dtype, [trait], [expr])
        
        return creator
    
    def add_creature(self, cnxn, genus, creator, species, trait, expr):
        '''
        Get or create a creature of a given creator.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        genus : str
            Name of the creator (parent) table.
        creator : pandas.DataFrame
            A single-row dataframe of creator entity data.
        species : str
            Name of the creature (child) table to be retrieved 
            or created.
        trait : str
            Name of the creature's text attribute.
        expr : str
            Value of the creature's text attribute.

        Returns
        -------
        creature : pandas.DataFrame
            A single-row dataframe of creature entity data.

        '''
        creator_id = creator.id.values[0]
        _, dtype = Tabula.get_creature_table(genus, species, trait)
        
        creature = self.get_or_create(
            cnxn, species, dtype, [trait, f'{genus}_id'], [expr, creator_id]
        )
        
        return creature
    
    
    ###########
    # getters #
    ###########
    
    def get_all_tables(self, cnxn):
        '''
        Get all tables in the instance database.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.

        Returns
        -------
        tables : pandas.DataFrame
            Dataframe describing all database tables.

        '''
        where = Inquiry.where(['type'], ['table'])
        where = f"{where} AND name NOT LIKE 'sqlite_%'"
        sql = Inquiry.select('sqlite_schema', ['name'], where)
        tables = list(pd.read_sql(sql, cnxn).name)
        
        return tables
    
    def get_table_columns(self, cnxn, tablename):
        '''
        Get columns of a database table.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        tablename : str
            Name of the database table.

        Returns
        -------
        columns : pandas.DataFrame
            Dataframe describing table columns.

        '''
        pragma = f'PRAGMA TABLE_INFO({tablename});'
        
        columns = pd.read_sql(pragma, cnxn)[
            ['name', 'type', 'notnull', 'pk']
        ].rename(columns={
            'name': 'column_name',
            'type': 'data_type',
            'notnull': 'not_null',
            'pk': 'is_pk'
        })
        
        return columns
    
    def get_df(self, cnxn, sql, expected_columns=None, raise_errors=False):
        '''
        Get result of SQL query as a pandas dataframe.
        In the event of an exception, return an empty
        dataframe.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            Connection to the database.
        sql : str
            SQL to be executed by pandas.read_sql.
        expected_columns : list(str), optional
            List of expected columns. If raise_errors is True 
            & the dataframe columns do not match expected_columns, 
            a ValueError is raised. The default is None.
        raise_errors : bool, optional
            Whether to raise exceptions encountered during 
            execution. The default is False.

        Raises
        ------
        error
            If raise_errors is True, raise any error encountered 
            during execution.

        Returns
        -------
        df : pandas.DataFrame
            A dataframe containing the results of the 
            SQL query.

        '''
        error = None
        
        try:
            df = pd.read_sql(sql, cnxn)
            
            if expected_columns and set(df.columns) != set(expected_columns):
                err_msg = ' '.join([
                    f'expected columns {expected_columns}.',
                    f'found {list(df.columns)}'
                ])
                
                error = ValueError(err_msg)
        
        except Exception as err:
            error = err
            df = pd.DataFrame(columns=expected_columns)
            
        if error and raise_errors:
            raise error
            
        return df
    
    def get_by_id(self, cnxn, tablename, entity_id):
        '''
        Get an entity by its id.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        tablename : str
            Name of the table to query.
        entity_id : str or int
            Value of the entity's id.

        Returns
        -------
        result : pandas.DataFrame
            A dataframe containing the query results.

        '''
        where = Inquiry.where(['id'], [entity_id])
        select = Inquiry.select(tablename, ['*'], where)
        result = pd.read_sql(select, cnxn)
        
        return result
    
    def get_creator_genus(self, cnxn, species):
        '''
        Get table name of creator (parent) table.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        species : str
            Name of the creature (child) table.

        Raises
        ------
        ValueError
            If more than one creator table is found, 
            raise a ValueError. Each creature table 
            should have one & only one creator.

        Returns
        -------
        genus : str
            Name of the creator (parent) table.

        '''
        pragma = f'PRAGMA FOREIGN_KEY_LIST({species});'
        foreign_keys = pd.read_sql(pragma, cnxn)
        genus = None
        
        if foreign_keys.shape[0] > 1:
            msg = ' '.join([
                'Data integrity error:',
                f'{species} shows more than one creator',
                f'({str(list(foreign_keys.table))}).'
            ])
            
            raise ValueError(msg)
            
        elif not foreign_keys.empty:
            genus = foreign_keys.table.values[0]
            
        return genus
    
    def get_creature_species(self, cnxn, genus):
        '''
        Get types of all creatures with a given creator genus.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        genus : str
            Name of the creator (parent) table.

        Returns
        -------
        species : list(str)
            List of creature (child) table names.

        '''
        tables = self.get_all_tables(cnxn)
        
        species = [
            t for t in tables if self.get_creator_genus(cnxn, t) == genus
        ]
        
        return species
                
    def get_creator(self, cnxn, species, creature):
        '''
        Get the creator of a given creature.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        species : str
            Name of the creature (child) table.
        creature : pandas.DataFrame
            A single-row dataframe of creature entity data.

        Returns
        -------
        genus : str
            Name of the creator (parent) table.
        creator : pandas.DataFrame
            A single-row dataframe of creator entity data.

        '''
        genus = self.get_creator_genus(cnxn, species)
        creator = []
        
        if genus:
            creator_id = creature[f'{genus}_id'].values[0]
            where = Inquiry.where(['id'], [creator_id])
            sql = Inquiry.select(genus, ['*'], where)
            creator = [(genus, pd.read_sql(sql, cnxn))]
        
        return creator
    
    def get_creatures(self, cnxn, genus, creator):
        '''
        Get all creatures of a given creator.

        Parameters
        ----------
        cnxn : sqlite3.Connection
            A database connection.
        genus : str
            Name of the creator (parent) table.
        creator : pandas.DataFrame
            A single-row dataframe of creator entity data.

        Returns
        -------
        creatures : list(tuple(str, pandas.DataFrame))
            List of two-tuples whose first entry is the 
            name of the creature (child) table, & whose 
            second entry is a dataframe of creature data.

        '''
        creator_id = creator.id.values[0]
        species = self.get_creature_species(cnxn, genus)
        creatures = []
        
        for s in species:
            where = Inquiry.where([f'{genus}_id'], [creator_id])
            sql = Inquiry.select(s, ['*'], where)
            
            members = pd.read_sql(sql, cnxn).apply(
                pd.to_numeric, errors='ignore'
            )
            
            creatures += [(s, members)]
            
        return creatures
            
    


from indexia.indexia import Indexia
from indexia.inquiry import Tabula
import os
import pandas as pd
import sqlite3
import unittest as ut


class TestIndexia(ut.TestCase):
    @classmethod        
    def setUpClass(cls):        
        cls.test_db = 'tests/data/test_indexia.db'
        
    @classmethod
    def getTestObjects(cls):
        creator = Tabula.get_creator_table(
            'creator', 'name'
        )
        
        creature = Tabula.get_creature_table(
            'creator', 'creature', 'name'
        )
        
        return creator, creature

    def setUp(self):
        creator, creature = self.getTestObjects()
        self.creator_table, self.creator_dtype = creator
        self.creature_table, self.creature_dtype = creature
        self.trait = 'name'
        self.creator_expr = 'father'
        self.creature_expr = 'son'
        
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            
            self.creator_data = ix.get_or_create(
                cnxn, self.creator_table, self.creator_dtype, 
                [self.trait], 
                [self.creator_expr]
            )
            
            self.creator_id, _ = self.creator_data.iloc[0]
            
            self.creature_data = ix.get_or_create(
                cnxn, self.creature_table, self.creature_dtype, 
                [self.trait, 'creator_id'], 
                [self.creature_expr, self.creator_id]
            )
            
            self.creature_id, _, _ = self.creature_data.iloc[0]
        
    def testOpenCnxn(self):
        with Indexia(self.test_db) as ix:
            cnxn_1 = ix.open_cnxn(ix.db)
            cnxn_2 = ix.open_cnxn(ix.db)
            
            self.assertEqual(len(ix.cnxns[self.test_db]), 2)
            self.assertIsInstance(cnxn_1, sqlite3.Connection)
            self.assertIsInstance(cnxn_2, sqlite3.Connection)
    
    def testCloseCnxn(self):
        with Indexia(self.test_db) as ix:
            ix.open_cnxn(ix.db)
            self.assertEqual(len(ix.cnxns[self.test_db]), 1)
            ix.close_cnxn(self.test_db)
            self.assertEqual(len(ix.cnxns[self.test_db]), 0)
    
    def testCloseAllCnxns(self):
        with Indexia(self.test_db) as ix:
            ix.open_cnxn(ix.db)
            self.assertEqual(len(ix.cnxns[self.test_db]), 1)
            ix.close_all_cnxns()
            
            for db in ix.cnxns:
                self.assertEqual(len(ix.cnxns[db]), 0)
                
    def testGetOrCreate(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            creator_expr = 'neonymos'
            
            self.assertRaises(
                ValueError, ix.get_or_create, 
                cnxn, self.creator_table, self.creator_dtype, 
                [self.trait], [creator_expr], retry=False
            )
            
            creator_data = ix.get_or_create(
                cnxn, self.creator_table, self.creator_dtype, 
                [self.trait], [creator_expr], retry=True
            )
            
            self.assertIsInstance(creator_data, pd.DataFrame),
            self.assertEqual(creator_data.shape[0], 1)
     
    def testDelete(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            deleted = ix.delete(cnxn, self.creator_table, self.creator_id)
            self.assertEqual(self.creator_id, deleted)
            
            self.assertRaises(
                ValueError, ix.get_or_create, 
                cnxn, self.creator_table, self.creator_dtype, 
                [self.trait], [self.creator_expr], retry=False
            )
            
    def testUpdate(self):        
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            
            rows_updated = ix.update(
                cnxn, self.creator_table, 
                [self.trait], ['pater'], 
                [self.trait], [self.creator_expr]
            )
                        
            updated = ix.get_or_create(
                cnxn, self.creator_table, self.creator_dtype, 
                ['id'], [self.creator_id]
            )
            
            self.assertEqual(rows_updated, 1)
            self.assertEqual(updated.loc[0, 'name'], 'pater')
            
    def testAddCreator(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            
            creator_data = ix.add_creator(
                cnxn, self.creator_table, self.trait, self.creator_expr
            )
            
            pd.testing.assert_frame_equal(creator_data, self.creator_data)
            
            creator_data = ix.add_creator(
                cnxn, self.creator_table, self.trait, 'neonymos'
            )
            
            creator_id, creator_expr = creator_data.iloc[0]
            self.assertEqual(creator_id, self.creator_id + 1)
            self.assertEqual(creator_expr, 'neonymos')
    
    def testAddCreature(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            
            creature_data = ix.add_creature(
                cnxn, self.creator_table, self.creator_data, 
                self.creature_table, self.trait, self.creature_expr
            )
            
            pd.testing.assert_frame_equal(creature_data, self.creature_data)
            
            creature_data = ix.add_creature(
                cnxn, self.creator_table, self.creator_data,
                self.creature_table, self.trait, 'neonymos'
            )
            
            creature_id, creature_expr, creator_id = creature_data.iloc[0]
            self.assertEqual(creator_id, self.creator_id)
            self.assertEqual(creature_id, self.creature_id + 1)
            self.assertEqual(creature_expr, 'neonymos')
                
    def testGetAllTables(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            table_list = ix.get_all_tables(cnxn)
            
            self.assertListEqual(
                table_list, [self.creator_table, self.creature_table]
            )
    
    def testGetTableColumns(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            column_data = ix.get_table_columns(cnxn, self.creator_table)
            
            exp_column_data = pd.DataFrame(data={
                'column_name': ['id', 'name'],
                'data_type': ['INTEGER', 'TEXT'],
                'not_null': [1, 1],
                'is_pk': [1, 0]
            })
            
            pd.testing.assert_frame_equal(column_data, exp_column_data)
    
    def testGetDF(self):
        creator_cols = ['id', 'name']
        valid_sql = f'SELECT * FROM {self.creator_table};'
        invalid_sql = 'SELECT * FROM nonexistent_table;'
        
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            
            expected_columns = None
            raise_errors = False
            df = ix.get_df(cnxn, valid_sql, expected_columns, raise_errors)
            self.assertIsInstance(df, pd.DataFrame)
            self.assertEqual(list(df.columns), creator_cols)
            
            df = ix.get_df(cnxn, invalid_sql, expected_columns, raise_errors)
            self.assertIsInstance(df, pd.DataFrame)
            self.assertEqual(list(df.columns), [])
            
            expected_columns = None
            raise_errors = True
            df = ix.get_df(cnxn, valid_sql, expected_columns, raise_errors)
            self.assertIsInstance(df, pd.DataFrame)
            self.assertEqual(list(df.columns), creator_cols)
            
            self.assertRaises(
                pd.io.sql.DatabaseError, ix.get_df, 
                cnxn, invalid_sql, expected_columns, raise_errors
            )
            
            expected_columns = ['invalid_column']
            raise_errors = False
            df = ix.get_df(cnxn, valid_sql, expected_columns, raise_errors)
            self.assertEqual(list(df.columns), creator_cols)
            
            expected_columns = ['invalid_column']
            raise_errors = True
            
            self.assertRaises(
                ValueError, ix.get_df, 
                cnxn, valid_sql, expected_columns, raise_errors
            )
            
            expected_columns = creator_cols
            raise_errors = True
            df = ix.get_df(cnxn, valid_sql, expected_columns, raise_errors)
            self.assertEqual(list(df.columns), creator_cols)
            self.assertGreaterEqual(df.shape[0], 1)    
    
    def testGetByID(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            
            creator_retrieved = ix.get_by_id(
                cnxn, self.creator_table, self.creator_id
            )
            
            pd.testing.assert_frame_equal(self.creator_data, creator_retrieved)
            
            expect_empty = ix.get_by_id(
                cnxn, self.creator_table, self.creator_id + 1
            )
            
            self.assertTrue(expect_empty.empty)
    
    def testGetCreatorGenus(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            genus = ix.get_creator_genus(cnxn, self.creature_table)
            self.assertEqual(self.creator_table, genus)
            
    def testGetCreatureSpecies(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            species = ix.get_creature_species(cnxn, self.creator_table)
            exp_species = ['creature']
            self.assertEqual(species, exp_species)
            
            species = ix.get_creature_species(cnxn, self.creature_table)
            exp_species = []
            self.assertEqual(species, exp_species)
    
    def testGetCreator(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)
            
            creator_genus, creator_data = ix.get_creator(
                cnxn, self.creature_table, self.creature_data
            )[0]
            
            exp_genus = ix.get_creator_genus(cnxn, self.creature_table)
            self.assertEqual(creator_genus, exp_genus)
            pd.testing.assert_frame_equal(creator_data, self.creator_data)
            
            expect_empty = ix.get_creator(
                cnxn, self.creator_table, self.creator_data
            )
            
            self.assertEqual(len(expect_empty), 0)
            
    def testGetCreatures(self):
        with Indexia(self.test_db) as ix:
            cnxn = ix.open_cnxn(ix.db)  
            
            creatures = ix.get_creatures(
                cnxn, self.creator_table, self.creator_data
            )
            
            exp_creatures = [('creature', pd.DataFrame(
                data={'id': [1], 'name': ['son'], 'creator_id': [1]}
            ))]
            
            species, members = creatures[0]
            exp_species, exp_members = exp_creatures[0]
            self.assertEqual(species, exp_species)
            pd.testing.assert_frame_equal(members, exp_members)
            
            exp_empty = ix.get_creatures(
                cnxn, self.creature_table, self.creature_data
            )
            
            self.assertEqual(len(exp_empty), 0)
    
    def tearDown(self):
        try:
            os.remove(self.test_db)
        except:
            pass


if __name__ == '__main__':
    ut.main()